/*******************************************************************************
 * Copyright (c) 2008 Mateusz Matela and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     Mateusz Matela - initial API and implementation
 *******************************************************************************/
package org.gsoc.eclipse.tostringgenerator;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

/**
 * <p>
 * A class used for parsing the toString() template and storing the results.
 * </p>
 * <p>
 * The template is divided into four parts:
 * <ul>
 * <li><b>Beginning</b> - all the characters from the beginning of the template
 * (inclusively) to the first occurrence of any member related token
 * (exclusively). This part can contain only object related tokens and is used
 * to generate the beginning of the generated <code>String</code>.</li>
 * <li><b>Body</b> - all the characters from the first occurrence of any member
 * related token to the token directly preceding the
 * <code>${otherMembers}</code> token, inclusively. The content of this part is
 * used once for every processed member. All member related and object related
 * tokens are available.
 * <li><b>Separator</b></li> - all the characters between a token directly
 * preceding the <code>${otherMembers}</code> token and the
 * <code>${otherMembers}</code>, exclusively. This part does not contain tokens
 * and is used to separate members in the generated <code>String</code>.
 * <li><b>Ending</b> - all the characters from the occurrence of the
 * <code>${otherMembers}</code> token (exclusively) to the end of the template
 * (inclusively). This part can also contain only object related tokens.</li>
 * </ul>
 * </p>
 * <p>
 * Note: for the template to work properly, it must contain at least one member
 * related token and exactly one ${otherMembers} token. Additionally, no member
 * related token can follow the ${otherMembers} token.
 * </p>
 */
public class DefaultToStringTemplateParser implements IToStringTemplateParser {
	protected String[] beginning;

	protected String[] body;

	protected String separator;

	protected String[] ending;

	protected Map descriptions;

	/**
	 * This method is used in {@link #parseTemplate(String)}. It can be
	 * overridden in subclasses to easily add new tokens.
	 * 
	 * @return member related tokens recognized by this parser
	 */
	protected String[] getMemberRelatedTokens() {
		return MEMBER_RELATED_TOKENS;
	}

	/**
	 * This method is used in {@link #parseTemplate(String)}. It can be
	 * overridden in subclasses to easily add new tokens.
	 * 
	 * @return object related tokens recognized by this parser
	 */
	protected String[] getObjectRelatedTokens() {
		return OBJECT_RELAED_TOKENS;
	}

	/**
	 * This method is used in {@link #parseTemplate(String)}. It can be
	 * overridden in subclasses to easily add new tokens.
	 * 
	 * @return member and object related tokens recognized by this parser (all
	 *         tokens but {$otherMembers})
	 */
	protected String[] getObjectAndMemberRelatedTokens() {
		return OBJECT_AND_MEMBER_RELATED_TOKENS;
	}

	public void parseTemplate(String template) {
		String[] emptyArray = new String[0];
		int beginningEnd = firstOccuranceOf(template, getMemberRelatedTokens());
		if (beginningEnd >= 0) {
			beginning = (String[]) extractElements(
					template.substring(0, beginningEnd),
					getObjectRelatedTokens()).toArray(emptyArray);
		} else {
			beginningEnd = 0;
			beginning = emptyArray;
		}
		int endingStart = template.indexOf(OTHER_MEMBERS_TOKEN);
		if (endingStart == -1)
			endingStart = template.length();

		ArrayList bodyList = extractElements(template.substring(beginningEnd,
				endingStart), getObjectAndMemberRelatedTokens());

		try {
			separator = (String) bodyList.get(bodyList.size() - 1);
			bodyList.remove(bodyList.size() - 1);
		} catch (Exception e) {
			separator = ""; //$NON-NLS-1$
		}

		body = (String[]) bodyList.toArray(emptyArray);
		ending = (String[]) extractElements(
				template.substring(endingStart + OTHER_MEMBERS_TOKEN.length()),
				getObjectRelatedTokens()).toArray(emptyArray);

	}

	protected int firstOccuranceOf(String template, String[] memberRelatedTokens) {
		int result = -1;
		for (int i = 0; i < memberRelatedTokens.length; i++) {
			int indexOf = template.indexOf(memberRelatedTokens[i]);
			if (result == -1 || (indexOf > 0 && indexOf < result))
				result = indexOf;
		}
		return result;
	}

	protected ArrayList extractElements(String template, String[] wantedTokens) {
		ArrayList result = new ArrayList();
		while (true) {
			if (template.length() == 0)
				break;
			String foundToken = null;
			int tokenPosition = template.length();
			for (int i = 0; i < wantedTokens.length; i++) {
				int position = template.indexOf(wantedTokens[i]);
				if (position >= 0 && position < tokenPosition) {
					tokenPosition = position;
					foundToken = wantedTokens[i];
				}
			}
			if (tokenPosition == template.length()) {
				result.add(template);
				break;
			}
			if (tokenPosition != 0)
				result.add(template.substring(0, tokenPosition));
			result.add(foundToken);
			template = template.substring(tokenPosition + foundToken.length());
		}
		return result;
	}

	public String[] getBeginning() {
		return beginning;
	}

	public String[] getBody() {
		return body;
	}

	public String[] getEnding() {
		return ending;
	}

	public String getSeparator() {
		return separator;
	}

	public Map getTokenDescriptions() {
		if (descriptions == null) {
			descriptions = new HashMap();
			for (int i = 0; i < tokens.length; i++)
				descriptions.put(tokens[i], tokenDescriptions[i]);
		}
		return descriptions;
	}

	public String[] getTokens() {
		return tokens;
	}
}
