/*******************************************************************************
 * Copyright (c) 2008 Mateusz Matela and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     Mateusz Matela - initial API and implementation
 *******************************************************************************/
package org.gsoc.eclipse.tostringgenerator;

import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.rewrite.ASTRewrite;
import org.eclipse.jdt.core.dom.rewrite.ImportRewrite;

/**
 * Classes implementing this interface provide all information necessary for
 * toString() method generation.
 */
public interface IToStringGenerationContext {
	/**
	 * 
	 * @return the binding of the type that toString() is generated for
	 */
	public ITypeBinding getTypeBinding();

	/**
	 * Returns the members that should be included in the toString() method.
	 * 
	 * @return an array containing
	 *         {@link org.eclipse.jdt.core.dom.IVariableBinding}s and/or
	 *         {@link org.eclipse.jdt.core.dom.IMethodBinding}
	 */
	public Object[] getSelectedMembers();

	/**
	 * 
	 * @return the template parser containing information about used template.
	 */
	public IToStringTemplateParser getTemplateParser();

	/**
	 * 
	 * @return the AST rewrite for the generator to write changes into
	 */
	public ASTRewrite getASTRewrite();

	/**
	 * 
	 * @return the AST that should own the generated code.
	 */
	public AST getAST();

	/**
	 * 
	 * @return the compilation unit of the type that toString() is generated for
	 */
	public ICompilationUnit getCompilationUnit();

	/**
	 * 
	 * @return the import rewrite used to add imports in modified code
	 */
	public ImportRewrite getImportRewrite();

	/**
	 * 
	 * @return true, if the toString() method should skip members that have null
	 *         value
	 */
	public boolean isSkipNulls();

	/**
	 * 
	 * @return true, if the toString() method should use custom methods to print
	 *         contents of arrays
	 */
	public boolean isCustomArray();

	/**
	 * 
	 * @return true, if the toString() method should should limit number of
	 *         items listed for arrays, Lists, Collections and Maps
	 */
	public boolean isLimitItems();

	/**
	 * This value makes sense only i {@link #isLimitItems()} returns true.
	 * 
	 * @return the maximum number of items that should be listed for arrays,
	 *         Lists, Collections and Maps
	 */
	public int getLimitItemsValue();

	/**
	 * 
	 * @return true, if a comment should be created for the toString() method
	 */
	public boolean isCreateComments();

	/**
	 * 
	 * @return true, if an @Override annotation should be added to the
	 *         toString() method
	 */
	public boolean isOverrideAnnotation();

	/**
	 * 
	 * @return true, if every access to a member should be preceded with 'this'
	 *         expression
	 */
	public boolean isKeywordThis();

	/**
	 * 
	 * @return true, if blocks should be forced for single statements in
	 *         if-then/for/while
	 */
	public boolean isForceBlocks();

	/**
	 * 
	 * @return true, if JDK1.5 API can be used by the toString() method
	 */
	public boolean is50orHigher();

	/**
	 * 
	 * @return true, if JDK1.6 API can be used by the toString() method
	 */
	public boolean is60orHigher();
}
