/*******************************************************************************
 * Copyright (c) 2008 Mateusz Matela and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     Mateusz Matela - initial API and implementation
 *******************************************************************************/
package org.gsoc.eclipse.tostringgenerator.internal;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtensionRegistry;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.InvalidRegistryObjectException;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.ISchedulingRule;

import org.eclipse.core.resources.IWorkspaceRunnable;
import org.eclipse.core.resources.ResourcesPlugin;

import org.eclipse.jface.dialogs.MessageDialog;

import org.eclipse.ltk.core.refactoring.RefactoringStatus;

import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IMember;
import org.eclipse.jdt.core.ISourceRange;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTMatcher;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;
import org.eclipse.jdt.core.dom.BodyDeclaration;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.rewrite.ASTRewrite;
import org.eclipse.jdt.core.dom.rewrite.ImportRewrite;
import org.eclipse.jdt.core.dom.rewrite.ListRewrite;

import org.eclipse.jdt.internal.corext.dom.ASTNodes;
import org.eclipse.jdt.internal.corext.refactoring.structure.CompilationUnitRewrite;
import org.eclipse.jdt.internal.corext.util.JavaModelUtil;
import org.eclipse.jdt.internal.corext.util.Messages;

import org.gsoc.eclipse.tostringgenerator.Activator;
import org.gsoc.eclipse.tostringgenerator.DefaultToStringTemplateParser;
import org.gsoc.eclipse.tostringgenerator.IToStringGenerationContext;
import org.gsoc.eclipse.tostringgenerator.IToStringGenerator;
import org.gsoc.eclipse.tostringgenerator.IToStringTemplateParser;

import org.eclipse.jdt.internal.ui.JavaPlugin;
import org.gsoc.eclipse.tostringgenerator.internal.dialogs.GenerateToStringDialog;

/**
 * <p>
 * A workspace runnable to add implementation for
 * <code>{@link java.lang.Object#toString()}</code>
 * </p>
 */
public class GenerateToStringOperation implements IWorkspaceRunnable {

	private static class ToStringGenerationContext implements
			IToStringGenerationContext {

		private Object[] fSelectedMembers;

		private IToStringTemplateParser fParser;

		private ITypeBinding fType;

		private GenerateToStringDialog.ToStringGenerationSettings fSettings;

		private CompilationUnitRewrite fRewrite;

		private ToStringGenerationContext(IToStringTemplateParser parser,
				Object[] selectedMembers,
				GenerateToStringDialog.ToStringGenerationSettings settings,
				ITypeBinding type, CompilationUnitRewrite rewrite) {
			fParser = parser;
			fSelectedMembers = selectedMembers;
			fSettings = settings;
			fType = type;
			fRewrite = rewrite;
		}

		public ASTRewrite getASTRewrite() {
			return fRewrite.getASTRewrite();
		}

		public AST getAST() {
			return fRewrite.getAST();
		}

		public ICompilationUnit getCompilationUnit() {
			return fRewrite.getCu();
		}

		public ImportRewrite getImportRewrite() {
			return fRewrite.getImportRewrite();
		}

		public int getLimitItemsValue() {
			return fSettings.limitValue;
		}

		public Object[] getSelectedMembers() {
			return fSelectedMembers;
		}

		public IToStringTemplateParser getTemplateParser() {
			return fParser;
		}

		public ITypeBinding getTypeBinding() {
			return fType;
		}

		public boolean is50orHigher() {
			return fSettings.is50orHigher;
		}

		public boolean is60orHigher() {
			return fSettings.is60orHigher;
		}

		public boolean isCreateComments() {
			return fSettings.createComments;
		}

		public boolean isCustomArray() {
			return fSettings.customArrayToString;
		}

		public boolean isForceBlocks() {
			return fSettings.useBlocks;
		}

		public boolean isKeywordThis() {
			return fSettings.useKeywordThis;
		}

		public boolean isLimitItems() {
			return fSettings.limitElements;
		}

		public boolean isOverrideAnnotation() {
			return fSettings.overrideAnnotation;
		}

		public boolean isSkipNulls() {
			return fSettings.skipNulls;
		}

	}

	/** The insertion point, or <code>null</code> */
	private IJavaElement fInsert;

	private CompilationUnitRewrite fRewrite;

	private IToStringGenerationContext fContext;

	private IToStringGenerator fGenerator;

	private CompilationUnit fUnit;

	private GenerateToStringOperation(IJavaElement insert,
			IToStringGenerationContext context, IToStringGenerator generator,
			CompilationUnit unit, CompilationUnitRewrite rewrite) {
		fInsert = insert;
		fContext = context;
		fRewrite = rewrite;
		fUnit = unit;
		fGenerator = generator;
	}

	public void run(IProgressMonitor monitor) throws CoreException {
		if (monitor == null)
			monitor = new NullProgressMonitor();
		try {
			monitor.beginTask("", 1); //$NON-NLS-1$
			monitor.setTaskName(UIMessages.GenerateToStringOperation_taskName);

			AbstractTypeDeclaration declaration = (AbstractTypeDeclaration) ASTNodes
					.findDeclaration(fContext.getTypeBinding(), fRewrite
							.getRoot());
			ListRewrite rewriter = fRewrite.getASTRewrite().getListRewrite(
					declaration, declaration.getBodyDeclarationsProperty());
			if (fContext.getTypeBinding() != null && rewriter != null) {

				MethodDeclaration toStringMethod = fGenerator
						.generateToStringMethod();

				final List list = (List) declaration
						.getStructuralProperty(declaration
								.getBodyDeclarationsProperty());
				BodyDeclaration replace = findMethodToReplace(list,
						toStringMethod);
				if (replace == null
						|| ((Boolean) toStringMethod
								.getProperty(IToStringGenerator.OVERWRITE_METHOD_PROPERTY))
								.booleanValue())
					insertMethod(toStringMethod, rewriter, replace);

				List helperMethods = fGenerator.generateHelperMethods();
				for (Iterator iterator = helperMethods.iterator(); iterator
						.hasNext();) {
					MethodDeclaration method = (MethodDeclaration) iterator
							.next();
					replace = findMethodToReplace(list, method);
					if (replace == null
							|| ((Boolean) method
									.getProperty(IToStringGenerator.OVERWRITE_METHOD_PROPERTY))
									.booleanValue()) {
						insertMethod(method, rewriter, replace);
					}
				}

				JavaModelUtil.applyEdit((ICompilationUnit) fUnit
						.getJavaElement(), fRewrite.createChange().getEdit(),
						false, monitor);
			}

		} finally {
			monitor.done();
		}
	}

	/**
	 * Generates toString() method and helper methods if necessary and returns
	 * them as a string.
	 * 
	 * @param importRewrite
	 *            the import rewrite that new imports should be added to
	 * 
	 * @return a string representing generated methods
	 */
	public String getMethodsAsString(ImportRewrite importRewrite) {
		try {
			StringBuffer buffer = new StringBuffer(fGenerator
					.generateToStringMethod().toString());
			List helperMethods = fGenerator.generateHelperMethods();
			for (Iterator iterator = helperMethods.iterator(); iterator
					.hasNext();) {
				MethodDeclaration method = (MethodDeclaration) iterator.next();
				buffer.append(method.toString());
			}
			String[] addedImports = fRewrite.getImportRewrite()
					.getAddedImports();
			for (int i = 0; i < addedImports.length; i++) {
				importRewrite.addImport(addedImports[i]);
			}
			addedImports = fRewrite.getImportRewrite().getAddedStaticImports();
			for (int i = 0; i < addedImports.length; i++) {
				int dotpos = addedImports[i].lastIndexOf('.');
				importRewrite.addStaticImport(addedImports[i].substring(0,
						dotpos), addedImports[i].substring(dotpos + 1), true);
				importRewrite.addStaticImport(addedImports[i].substring(0,
						dotpos), addedImports[i].substring(dotpos + 1), false);
			}
			return buffer.toString();
		} catch (CoreException e) {
			return ""; //$NON-NLS-1$
		}
	}

	/**
	 * @return RefactoringStatus with eventual errors and warnings
	 */
	public RefactoringStatus checkConditions() {
		return fGenerator.checkConditions();
	}

	protected void insertMethod(MethodDeclaration method, ListRewrite rewriter,
			BodyDeclaration replace) throws JavaModelException {
		if (replace != null) {
			rewriter.replace(replace, method, null);
		} else {
			ASTNode insertion = getNodeToInsertBefore(rewriter, fInsert);
			if (insertion != null)
				rewriter.insertBefore(method, insertion, null);
			else
				rewriter.insertLast(method, null);
		}
	}

	/**
	 * Evaluates the insertion position of a new node.
	 * 
	 * @param listRewrite
	 *            The list rewriter to which the new node will be added
	 * @param sibling
	 *            The Java element before which the new element should be added.
	 * @return the AST node of the list to insert before or null to insert as
	 *         last.
	 * @throws JavaModelException
	 *             thrown if accessing the Java element failed
	 */

	private ASTNode getNodeToInsertBefore(ListRewrite listRewrite,
			IJavaElement sibling) throws JavaModelException {
		if (sibling instanceof IMember) {
			ISourceRange sourceRange = ((IMember) sibling).getSourceRange();
			if (sourceRange == null) {
				return null;
			}
			int insertPos = sourceRange.getOffset();

			List members = listRewrite.getOriginalList();
			for (int i = 0; i < members.size(); i++) {
				ASTNode curr = (ASTNode) members.get(i);
				if (curr.getStartPosition() >= insertPos) {
					return curr;
				}
			}
		}
		return null;
	}

	/**
	 * Determines if given method exists in a given list
	 * 
	 * @param list
	 *            list of method to search through
	 * @param method
	 *            method to find
	 * @return declaration of method from the list that has the same name and
	 *         parameter types, or null if not found
	 */
	protected BodyDeclaration findMethodToReplace(final List list,
			MethodDeclaration method) {
		for (final Iterator iterator = list.iterator(); iterator.hasNext();) {
			final BodyDeclaration bodyDecl = (BodyDeclaration) iterator.next();
			if (bodyDecl instanceof MethodDeclaration) {
				final MethodDeclaration method2 = (MethodDeclaration) bodyDecl;
				if (method2.getName().getIdentifier().equals(
						method.getName().getIdentifier())
						&& method2.parameters().size() == method.parameters()
								.size()) {
					Iterator iterator1 = method.parameters().iterator();
					Iterator iterator2 = method2.parameters().iterator();
					boolean ok = true;
					while (iterator1.hasNext()) {
						if (!((SingleVariableDeclaration) iterator1.next())
								.getType().subtreeMatch(
										new ASTMatcher(),
										((SingleVariableDeclaration) iterator2
												.next()).getType())) {
							ok = false;
							break;
						}
					}
					if (ok)
						return method2;
				}
			}
		}
		return null;
	}

	public ISchedulingRule getSchedulingRule() {
		return ResourcesPlugin.getWorkspace().getRoot();
	}

	private final static class ToStringGeneratorHandle {
		private final String NAME_ATTR = "name"; //$NON-NLS-1$

		private final String CLASS_ATTR = "class"; //$NON-NLS-1$

		private final String PARSER_CHILD = "toStringTemplateParser"; //$NON-NLS-1$

		private IConfigurationElement element;

		public ToStringGeneratorHandle(IConfigurationElement element) {
			this.element = element;
		}

		public String getName() {
			try {
				return element.getAttribute(NAME_ATTR);
			} catch (InvalidRegistryObjectException e) {
				elementInvalid(e);
			}
			return null;
		}

		public IToStringGenerator createGenerator() {
			try {
				return (IToStringGenerator) element
						.createExecutableExtension(CLASS_ATTR);
			} catch (InvalidRegistryObjectException e) {
				elementInvalid(e);
			} catch (CoreException e) {
				JavaPlugin.log(e.getStatus());
			}
			return createToStringGenerator(0);
		}

		public IToStringTemplateParser createParser() {
			try {
				IConfigurationElement[] parser = element
						.getChildren(PARSER_CHILD);
				if (parser.length > 0) {
					return (IToStringTemplateParser) parser[0]
							.createExecutableExtension(CLASS_ATTR);
				}
			} catch (InvalidRegistryObjectException e) {
				elementInvalid(e);
			} catch (CoreException e) {
				JavaPlugin.log(e.getStatus());
			}
			return new DefaultToStringTemplateParser();
		}

		private void elementInvalid(Throwable e) {
			Object[] args = { element.toString() };
			String message = Messages.format(
					UIMessages.GenerateToStringOperation_invalidExtension, args);
			IStatus status = new Status(IStatus.WARNING, JavaPlugin
					.getPluginId(), IStatus.OK, message, e);
			JavaPlugin.log(status);
			MessageDialog.openError(JavaPlugin.getActiveWorkbenchShell(),
					UIMessages.GenerateToStringOperation_errorDialogCaption, message);
		}
	}

	private static ToStringGeneratorHandle[] fHandles;

	private static String[] fStyleNames;

	private static final String EXTENSION_POINT = "javaToStringGenerator"; //$NON-NLS-1$

	public static final int STRING_CONCATENATION = 0;

	public static final int STRING_BUILDER = 1;

	public static final int STRING_BUILDER_CHAINED = 2;

	public static final int STRING_FORMAT = 3;

	public static final int APACHE_BUILDER = 4;

	public static final int APACHE_BUILDER_CHAINED = 5;

	public static final int SPRING_CREATOR = 6;

	public static final int SPRING_CREATOR_CHAINED = 7;

	public static final int ID_START = 8;

	private final static String[] hardcoded_styles = { UIMessages.GenerateToStringOperation_styleNameStringConcatenation,
			UIMessages.GenerateToStringOperation_styleNameStringBuilder,
			UIMessages.GenerateToStringOperation_styleNameStringBuilderChained,
			UIMessages.GenerateToStringOperation_styleNameStringFormat,
			UIMessages.GenerateToStringOperation_styleNameApache,
			UIMessages.GenerateToStringOperation_styleNameApacheChained,
			UIMessages.GenerateToStringOperation_styleNameSpring,
			UIMessages.GenerateToStringOperation_styleNameSpringChained };

	/**
	 * @return Array containing names of implemented code styles
	 */
	public static String[] getStyleNames() {
		return fStyleNames;
	}

	public static String[] refreshStyles() {
		IExtensionRegistry registry = Platform.getExtensionRegistry();
		IConfigurationElement[] elements = registry
				.getConfigurationElementsFor(Activator.PLUGIN_ID,
						EXTENSION_POINT);
		ArrayList handleList = new ArrayList();
		ArrayList nameList = new ArrayList();
		nameList.addAll(Arrays.asList(hardcoded_styles));
		for (int i = 0; i < elements.length; i++) {
			ToStringGeneratorHandle handle = new ToStringGeneratorHandle(
					elements[i]);
			String name = handle.getName();
			if (name != null) {
				nameList.add(name);
				handleList.add(handle);
			}
		}
		fStyleNames = (String[]) nameList.toArray(new String[0]);
		fHandles = (ToStringGeneratorHandle[]) handleList
				.toArray(new ToStringGeneratorHandle[0]);
		return fStyleNames;
	}

	/**
	 * 
	 * @param toStringStyle
	 *            id number of the code style (its position in the array
	 *            returned by {@link #getStyleNames()}
	 * @return a toString() generator implementing given code style
	 */
	private static IToStringGenerator createToStringGenerator(int toStringStyle) {
		switch (toStringStyle) {
		case STRING_CONCATENATION:
			return new StringConcatenationGenerator();
		case STRING_BUILDER:
			return new StringBuilderGenerator();
		case STRING_BUILDER_CHAINED:
			return new StringBuilderChainGenerator();
		case STRING_FORMAT:
			return new StringFormatGenerator();
		case APACHE_BUILDER:
			return new ApacheBuilderSpringCreatorGenerator(
					"org.apache.commons.lang.builder.ToStringBuilder", "builder", false); //$NON-NLS-1$ //$NON-NLS-2$ 
		case APACHE_BUILDER_CHAINED:
			return new ApacheBuilderSpringCreatorGenerator(
					"org.apache.commons.lang.builder.ToStringBuilder", "builder", true); //$NON-NLS-1$ //$NON-NLS-2$ 
		case SPRING_CREATOR:
			return new ApacheBuilderSpringCreatorGenerator(
					"org.springframework.core.style.ToStringCreator", "creator", false); //$NON-NLS-1$ //$NON-NLS-2$ 
		case SPRING_CREATOR_CHAINED:
			return new ApacheBuilderSpringCreatorGenerator(
					"org.springframework.core.style.ToStringCreator", "creator", true); //$NON-NLS-1$ //$NON-NLS-2$ 
		default:
			return fHandles[toStringStyle - ID_START].createGenerator();
		}
	}

	/**
	 * @param toStringStyle
	 *            id number of the style (its position in the array returned by
	 *            {@link #getStyleNames()}
	 * @return a template parser that should be used with given code style
	 */
	public static IToStringTemplateParser createTemplateParser(int toStringStyle) {
		if (toStringStyle >= ID_START) {
			return fHandles[toStringStyle - ID_START].createParser();
		}
		return new DefaultToStringTemplateParser();
	}

	/**
	 * Creates new <code>GenerateToStringOperation</code>, using
	 * <code>settings.toStringStyle</code> field to choose the right subclass.
	 * 
	 * @param typeBinding
	 * @param selectedBindings
	 * @param unit
	 * @param elementPosition
	 * @param settings
	 * @return a ready to use <code>GenerateToStringOperation</code> object
	 */
	public static GenerateToStringOperation createOperation(
			ITypeBinding typeBinding, Object[] selectedBindings,
			CompilationUnit unit, IJavaElement elementPosition,
			GenerateToStringDialog.ToStringGenerationSettings settings) {
		IToStringGenerator generator = createToStringGenerator(settings.toStringStyle);
		IToStringTemplateParser parser = createTemplateParser(settings.toStringStyle);
		parser.parseTemplate(settings.stringFormatTemplate);
		CompilationUnitRewrite rewrite = new CompilationUnitRewrite(
				(ICompilationUnit) unit.getJavaElement(), unit);
		IToStringGenerationContext context = new ToStringGenerationContext(
				parser, selectedBindings, settings, typeBinding, rewrite);
		generator.setContext(context);
		return new GenerateToStringOperation(elementPosition, context,
				generator, unit, rewrite);
	}

}
