/*******************************************************************************
 * Copyright (c) 2008 Mateusz Matela and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     Mateusz Matela - initial API and implementation
 *******************************************************************************/
package org.gsoc.eclipse.tostringgenerator.internal.actions;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.core.resources.IWorkspaceRunnable;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.jobs.ISchedulingRule;
import org.eclipse.jdt.core.IClassFile;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IMember;
import org.eclipse.jdt.core.IOpenable;
import org.eclipse.jdt.core.ISourceReference;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.internal.corext.codemanipulation.CodeGenerationSettings;
import org.eclipse.jdt.internal.corext.dom.ASTNodes;
import org.eclipse.jdt.internal.corext.dom.NodeFinder;
import org.eclipse.jdt.internal.corext.util.JavaModelUtil;
import org.eclipse.jdt.internal.corext.util.Messages;
import org.eclipse.jdt.internal.ui.JavaPlugin;
import org.eclipse.jdt.internal.ui.actions.SelectionConverter;
import org.eclipse.jdt.internal.ui.actions.WorkbenchRunnableAdapter;
import org.eclipse.jdt.internal.ui.javaeditor.CompilationUnitEditor;
import org.eclipse.jdt.internal.ui.preferences.JavaPreferencesSettings;
import org.eclipse.jdt.internal.ui.util.BusyIndicatorRunnableContext;
import org.eclipse.jdt.internal.ui.util.ElementValidator;
import org.eclipse.jdt.internal.ui.util.ExceptionHandler;
import org.eclipse.jdt.ui.JavaUI;
import org.eclipse.jdt.ui.PreferenceConstants;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.operation.IRunnableContext;
import org.eclipse.jface.text.IRewriteTarget;
import org.eclipse.jface.text.ITextSelection;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.window.Window;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;
import org.eclipse.ltk.core.refactoring.RefactoringStatusContext;
import org.eclipse.ltk.ui.refactoring.RefactoringUI;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IEditorActionDelegate;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IObjectActionDelegate;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.IWorkbenchWindowActionDelegate;
import org.eclipse.ui.PlatformUI;
import org.gsoc.eclipse.tostringgenerator.internal.GenerateToStringOperation;
import org.gsoc.eclipse.tostringgenerator.internal.dialogs.GenerateToStringDialog;
import org.gsoc.eclipse.tostringgenerator.internal.dialogs.GenerateToStringDialog.ToStringGenerationSettings;

/**
 * <p>
 * The delegate for the action that adds method implementations for
 * <code>{@link java.lang.Object#toString()}</code>. The action opens a dialog
 * from which the user can choose the fields and methods to be considered.
 * </p>
 * <p>
 * Will open the parent compilation unit in a Java editor. The result is
 * unsaved, so the user can decide if the changes are acceptable.
 * </p>
 * <p>
 * The action is applicable to structured selections containing elements of type
 * {@link org.eclipse.jdt.core.IType}.
 * </p>
 */
public class GenerateToStringActionDelegate implements IEditorActionDelegate,
		IWorkbenchWindowActionDelegate, IObjectActionDelegate {
	private static CompilationUnitEditor fEditor;
	private static CompilationUnit fUnit;
	private static ITypeBinding fTypeBinding;
	private static IWorkbenchWindow fWindow;
	private static IStructuredSelection fSelection;
	private static IAction fEditorAction, fSelectionAction;

	/**
	 * The constructor.
	 */
	public GenerateToStringActionDelegate() {
		// nothing to do
	}

	public void init(IWorkbenchWindow window) {
		fWindow = window;
	}

	public void dispose() {
		// nothing to dispose

	}

	public void selectionChanged(IAction action, ISelection selection) {
		fSelectionAction = action;
		if (selection instanceof IStructuredSelection) {
			fSelection = (IStructuredSelection) selection;
			setEnabled(canEnable(fSelection));
		} else {
			fSelection = null;
			if (selection instanceof ITextSelection
					&& (fEditor != null || (fWindow != null && fWindow
							.getActivePage().getActiveEditor() instanceof CompilationUnitEditor)))
				setEnabled(true);
			else
				setEnabled(false);
		}
	}

	public void setActiveEditor(IAction action, IEditorPart targetEditor) {
		fEditorAction = action;
		if (targetEditor instanceof CompilationUnitEditor) {
			fEditor = (CompilationUnitEditor) targetEditor;
			setEnabled(true);
		} else {
			fEditor = null;
			setEnabled(false);
		}
	}

	public void setActivePart(IAction action, IWorkbenchPart targetPart) {
		fWindow = targetPart.getSite().getWorkbenchWindow();
	}

	public void run(IAction action) {
		
		if (fSelection != null) {
			run(fSelection);
			return;
		}
		if (fWindow != null)
			try {
				fEditor = (CompilationUnitEditor) fWindow.getActivePage()
						.getActiveEditor();
			} catch (Exception e) {
				MessageDialog.openInformation(getShell(), getErrorCaption(),
						UIMessages.GenerateToStringActionDelegate_wrongEditor);
				return;
			}
		try {
			checkAndRun(SelectionConverter.getTypeAtOffset(fEditor));
		} catch (CoreException e) {
			ExceptionHandler.handle(e, getShell(), getErrorCaption(),
					UIMessages.GenerateToStringActionDelegate_cannotCreate);
		} catch (Exception e) {
			MessageDialog.openInformation(getShell(), getErrorCaption(),
					UIMessages.GenerateToStringActionDelegate_cannotCreate);
		}
	}

	private void run(IStructuredSelection selection) {
		try {
			checkAndRun(getSelectedType(selection));
		} catch (CoreException exception) {
			ExceptionHandler.handle(exception, getShell(), getErrorCaption(),
					UIMessages.GenerateToStringActionDelegate_cannotCreate);
		}
	}

	/**
	 * Returns the single selected type from the specified selection.
	 * 
	 * @param selection
	 *            the selection
	 * @return a single selected type, or <code>null</code>
	 * @throws JavaModelException
	 *             if the kind of the selection cannot be determined
	 */
	private IType getSelectedType(final IStructuredSelection selection)
			throws JavaModelException {
		if (selection.size() == 1
				&& selection.getFirstElement() instanceof IType) {
			final IType type = (IType) selection.getFirstElement();
			if (type.getCompilationUnit() != null && type.isClass())
				return type;
		} else if (selection.getFirstElement() instanceof ICompilationUnit) {
			final ICompilationUnit unit = (ICompilationUnit) selection
					.getFirstElement();
			final IType type = unit.findPrimaryType();
			if (type != null && type.isClass())
				return type;
		}
		return null;
	}

	/**
	 * Can this action be enabled on the specified selection?
	 * 
	 * @param selection
	 *            the selection to test
	 * @return <code>true</code> if it can be enabled, <code>false</code>
	 *         otherwise
	 */
	private boolean canEnable(final IStructuredSelection selection) {
		try {
			if (selection.size() == 1) {
				final Object element = selection.getFirstElement();
				if (element instanceof IType) {
					final IType type = (IType) element;
					return type.getCompilationUnit() != null && type.isClass();
				}
				if (element instanceof ICompilationUnit)
					return true;
			}
			if (selection.isEmpty()
					&& (fEditor != null || (fWindow != null && fWindow
							.getActivePage().getActiveEditor() instanceof CompilationUnitEditor))) {
				fSelection = null;
				return true;
			}
		} catch (Exception e) {
			return false;
		}
		return false;
	}

	private String getErrorCaption() {
		return UIMessages.GenerateToStringActionDelegate_errorDialogCaption;
	}

	private void setEnabled(boolean enable) {
		if (fEditorAction != null)
			fEditorAction.setEnabled(enable);
		if (fSelectionAction != null)
			fSelectionAction.setEnabled(enable);
	}

	private Shell getShell() {
		if (fWindow != null)
			return fWindow.getShell();
		return fEditor.getSite().getShell();
	}

	private void checkAndRun(IType type) throws CoreException {
		if (type == null) {
			MessageDialog.openInformation(getShell(), getErrorCaption(),
					UIMessages.GenerateToStringActionDelegate_wrongSelection);
			return;
		}
		if (!ElementValidator.check(type, getShell(), getErrorCaption(), false)
				|| !ActionUtil.isEditable(fEditor, getShell(), type)) {
			return;
		}
		if (type.isAnnotation()) {
			MessageDialog.openInformation(getShell(), getErrorCaption(),
					UIMessages.GenerateToStringActionDelegate_errorAnnotations);
			return;
		}
		if (type.isInterface()) {
			MessageDialog.openInformation(getShell(), getErrorCaption(),
					UIMessages.GenerateToStringActionDelegate_errorInterfaces);
			return;
		}
		if (type.isEnum()) {
			MessageDialog.openInformation(getShell(), getErrorCaption(),
					UIMessages.GenerateToStringActionDelegate_errorEnums);
			return;
		}
		if (type.isAnonymous()) {
			MessageDialog.openError(getShell(), getErrorCaption(),
					UIMessages.GenerateToStringActionDelegate_errorAnonumous);
			return;
		}
		run(getShell(), type);
	}

	/**
	 * Runs the action.
	 * 
	 * @param shell
	 *            the shell to use
	 * @param type
	 *            the type to generate stubs for
	 * @throws CoreException
	 *             if an error occurs
	 */
	protected void run(Shell shell, IType type) throws CoreException {

		initialize(type);

		boolean regenerate = false;
		if (isMethodAlreadyImplemented(fTypeBinding)) {
			regenerate = MessageDialog
					.openQuestion(
							getShell(),
							getErrorCaption(),
							Messages
									.format(
											UIMessages.GenerateToStringActionDelegate_replaceExistingQuestion,
											new String[] {
													fTypeBinding
															.getQualifiedName(),
													getToStringMethodName() }));
			if (!regenerate) {
				return;
			}
		}

		generateCandidates();

		final GenerateToStringDialog dialog = createDialog(shell, type);
		final int dialogResult = dialog.open();

		if (dialogResult == Window.OK) {

			final Object[] selected = dialog.getResult();
			if (selected == null) {
				return;
			}
			final ToStringGenerationSettings settings = createSettings(type,
					dialog);
			final IWorkspaceRunnable operation = createOperation(selected,
					settings, regenerate, type, dialog.getElementPosition());

			ITypeBinding superclass = fTypeBinding.getSuperclass();
			RefactoringStatus status = new RefactoringStatus();

			status.merge(checkGeneralConditions(type, settings, selected));

			if (!"java.lang.Object".equals(superclass.getQualifiedName())) { //$NON-NLS-1$
				status.merge(checkSuperClass(superclass));
			}

			for (int i = 0; i < selected.length; i++) {
				status.merge(checkMember(selected[i]));
			}

			if (status.hasEntries()) {
				Dialog d = RefactoringUI.createLightWeightStatusDialog(status,
						getShell(), getErrorCaption());
				if (d.open() != Window.OK) {
					return;
				}
			}

			final IEditorPart editor = JavaUI.openInEditor(type
					.getCompilationUnit());
			final IRewriteTarget target = editor != null ? (IRewriteTarget) editor
					.getAdapter(IRewriteTarget.class)
					: null;

			if (target != null)
				target.beginCompoundChange();
			try {
				IRunnableContext context = JavaPlugin
						.getActiveWorkbenchWindow();
				if (context == null)
					context = new BusyIndicatorRunnableContext();
				ISchedulingRule schedulingRule = ResourcesPlugin.getWorkspace()
						.getRoot();
				PlatformUI.getWorkbench().getProgressService()
						.runInUI(
								context,
								new WorkbenchRunnableAdapter(operation,
										schedulingRule), schedulingRule);
			} catch (InvocationTargetException exception) {
				ExceptionHandler.handle(exception, shell, getErrorCaption(),
						null);
			} catch (InterruptedException exception) {
				// Do nothing. Operation has been canceled by user.
			} finally {
				if (target != null)
					target.endCompoundChange();
			}
		}
	}

	private static final String METHODNAME_TO_STRING = "toString"; //$NON-NLS-1$

	private List fFields;

	private List fInheritedFields;

	private List fSelectedFields;

	private List fMethods;

	private List fInheritedMethods;

	private GenerateToStringOperation operation;

	private class ToStringInfo {

		public boolean foundToString = false;

		public boolean foundFinalToString = false;

		public ToStringInfo(ITypeBinding typeBinding) {
			IMethodBinding[] declaredMethods = typeBinding.getDeclaredMethods();

			for (int i = 0; i < declaredMethods.length; i++) {
				if (declaredMethods[i].getName().equals(METHODNAME_TO_STRING)
						&& declaredMethods[i].getParameterTypes().length == 0) {
					this.foundToString = true;
					if (Modifier.isFinal(declaredMethods[i].getModifiers()))
						this.foundFinalToString = true;
				}
			}
		}
	}

	protected RefactoringStatus checkMember(Object object) {
		return new RefactoringStatus();
	}

	RefactoringStatus checkGeneralConditions(IType type,
			ToStringGenerationSettings settings, Object[] selected) {
		return operation.checkConditions();
	}

	protected RefactoringStatus checkSuperClass(ITypeBinding superclass) {
		RefactoringStatus status = new RefactoringStatus();
		if (new ToStringInfo(superclass).foundFinalToString) {
			status
					.addError(
							Messages
									.format(
											UIMessages.GenerateToStringActionDelegate_errorFinalMethod,
											new String[] {
													Messages
															.format(
																	UIMessages.GenerateToStringActionDelegate_superClass,
																	superclass
																			.getQualifiedName()),
													getToStringMethodName() }),
							createRefactoringStatusContext(superclass
									.getJavaElement()));
		}
		return status;
	}

	protected static RefactoringStatusContext createRefactoringStatusContext(
			IJavaElement element) {
		if (element instanceof IMember) {
			return JavaStatusContext.create((IMember) element);
		}
		if (element instanceof ISourceReference) {
			IOpenable openable = element.getOpenable();
			try {
				if (openable instanceof ICompilationUnit) {
					return JavaStatusContext.create(
							(ICompilationUnit) openable,
							((ISourceReference) element).getSourceRange());
				} else if (openable instanceof IClassFile) {
					return JavaStatusContext.create((IClassFile) openable,
							((ISourceReference) element).getSourceRange());
				}
			} catch (JavaModelException e) {
				// ignore
			}
		}
		return null;
	}

	protected GenerateToStringDialog createDialog(Shell shell, IType type)
			throws JavaModelException {
		IVariableBinding[] fieldBindings = (IVariableBinding[]) fFields
				.toArray(new IVariableBinding[0]);
		IVariableBinding[] inheritedFieldBindings = (IVariableBinding[]) fInheritedFields
				.toArray(new IVariableBinding[0]);
		IVariableBinding[] selectedFieldBindings = (IVariableBinding[]) fSelectedFields
				.toArray(new IVariableBinding[0]);
		IMethodBinding[] methodBindings = (IMethodBinding[]) fMethods
				.toArray(new IMethodBinding[0]);
		IMethodBinding[] inheritededMethodBindings = (IMethodBinding[]) fInheritedMethods
				.toArray(new IMethodBinding[0]);
		return new GenerateToStringDialog(shell, fEditor, type, fieldBindings,
				inheritedFieldBindings, selectedFieldBindings, methodBindings,
				inheritededMethodBindings);
	}

	protected IWorkspaceRunnable createOperation(Object[] selectedBindings,
			ToStringGenerationSettings settings, boolean regenerate,
			IJavaElement type, IJavaElement elementPosition) {
		return operation = GenerateToStringOperation.createOperation(
				fTypeBinding, selectedBindings, fUnit, elementPosition,
				settings);
	}

	protected ToStringGenerationSettings createSettings(IType type,
			GenerateToStringDialog dialog) {
		ToStringGenerationSettings settings = dialog.getGenerationSettings();

		CodeGenerationSettings settings2 = JavaPreferencesSettings
				.getCodeGenerationSettings(type.getJavaProject());
		settings2.createComments = dialog.getGenerateComment();
		settings2.setSettings(settings);

		settings.createComments = dialog.getGenerateComment();
		settings.useBlocks = useBlocks(type.getJavaProject());
		String version = fUnit.getJavaElement().getJavaProject().getOption(
				JavaCore.COMPILER_SOURCE, true);
		settings.is50orHigher = !JavaModelUtil.isVersionLessThan(version,
				JavaCore.VERSION_1_5);
		settings.is60orHigher = !JavaModelUtil.isVersionLessThan(version,
				JavaCore.VERSION_1_6);
		return settings;
	}

	public static boolean useBlocks(IJavaProject project) {
		final String TRUE = "true"; //$NON-NLS-1$
		final String CONTROL_STATEMENTS_USE_BLOCKS = "cleanup.use_blocks"; //$NON-NLS-1$
		final String CONTROL_STATMENTS_USE_BLOCKS_ALWAYS = "cleanup.always_use_blocks"; //$NON-NLS-1$
		final String ONTROL_STATMENTS_USE_BLOCKS_NO_FOR_RETURN_AND_THROW = "cleanup.use_blocks_only_for_return_and_throw"; //$NON-NLS-1$
		if (TRUE.equals(PreferenceConstants.getPreference(
				CONTROL_STATEMENTS_USE_BLOCKS, project))) {
			return TRUE.equals(PreferenceConstants.getPreference(
					CONTROL_STATMENTS_USE_BLOCKS_ALWAYS, project))
					|| TRUE
							.equals(PreferenceConstants
									.getPreference(
											ONTROL_STATMENTS_USE_BLOCKS_NO_FOR_RETURN_AND_THROW,
											project));
		}
		return false;
	}

	protected boolean generateCandidates() {
		IVariableBinding[] candidateFields = fTypeBinding.getDeclaredFields();
		fFields = new ArrayList();
		fSelectedFields = new ArrayList();
		for (int i = 0; i < candidateFields.length; i++) {
			if (!Modifier.isStatic(candidateFields[i].getModifiers())) {
				fFields.add(candidateFields[i]);
				if (!Modifier.isTransient(candidateFields[i].getModifiers()))
					fSelectedFields.add(candidateFields[i]);
			}
		}

		IMethodBinding[] candidateMethods = fTypeBinding.getDeclaredMethods();
		fMethods = new ArrayList();
		for (int i = 0; i < candidateMethods.length; i++) {
			if (!Modifier.isStatic(candidateMethods[i].getModifiers())
					&& candidateMethods[i].getParameterTypes().length == 0
					&& !candidateMethods[i].getReturnType().getName().equals(
							"void") && !candidateMethods[i].getName().equals("toString") && !candidateMethods[i].getName().equals("clone")) { //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
				fMethods.add(candidateMethods[i]);
			}
		}

		fInheritedFields = new ArrayList();
		fInheritedMethods = new ArrayList();
		ITypeBinding typeBinding = fTypeBinding;
		while ((typeBinding = typeBinding.getSuperclass()) != null) {
			candidateFields = typeBinding.getDeclaredFields();
			for (int i = 0; i < candidateFields.length; i++) {
				if (!Modifier.isPrivate(candidateFields[i].getModifiers())
						&& !Modifier
								.isStatic(candidateFields[i].getModifiers())
						&& !contains(fFields, candidateFields[i])
						&& !contains(fInheritedFields, candidateFields[i])) {
					fInheritedFields.add(candidateFields[i]);
				}
			}
			candidateMethods = typeBinding.getDeclaredMethods();
			for (int i = 0; i < candidateMethods.length; i++) {
				if (!Modifier.isPrivate(candidateMethods[i].getModifiers())
						&& !Modifier.isStatic(candidateMethods[i]
								.getModifiers())
						&& candidateMethods[i].getParameterTypes().length == 0
						&& !candidateMethods[i].getReturnType().getName()
								.equals("void") && !contains(fMethods, candidateMethods[i]) && !contains(fInheritedMethods, candidateMethods[i]) && !candidateMethods[i].getName().equals("clone")) { //$NON-NLS-1$ //$NON-NLS-2$
					fInheritedMethods.add(candidateMethods[i]);
				}
			}
		}

		return true;
	}

	private boolean contains(List inheritedFields, Object member) {
		for (Iterator iterator = inheritedFields.iterator(); iterator.hasNext();) {
			Object object = iterator.next();
			if (object instanceof IVariableBinding
					&& member instanceof IVariableBinding)
				if (((IVariableBinding) object).getName().equals(
						((IVariableBinding) member).getName()))
					return true;
			if (object instanceof IMethodBinding
					&& member instanceof IMethodBinding)
				if (((IMethodBinding) object).getName().equals(
						((IMethodBinding) member).getName()))
					return true;
		}
		return false;
	}

	protected String getToStringMethodName() {
		return "toString()"; //$NON-NLS-1$
	}

	protected boolean isMethodAlreadyImplemented(ITypeBinding typeBinding) {
		return new ToStringInfo(typeBinding).foundToString;
	}

	protected void initialize(IType type) throws JavaModelException {
		fUnit = parse(type.getCompilationUnit());
		fTypeBinding = null;
		// type cannot be anonymous
		final AbstractTypeDeclaration declaration = (AbstractTypeDeclaration) ASTNodes
				.getParent(NodeFinder.perform(fUnit, type.getNameRange()),
						AbstractTypeDeclaration.class);
		if (declaration != null)
			fTypeBinding = declaration.resolveBinding();
	}

	private CompilationUnit parse(IJavaElement unit) {
		ASTParser fParser = ASTParser.newParser(AST.JLS3);
		fParser.setResolveBindings(true);
		fParser.setStatementsRecovery(false);
		if (unit instanceof ICompilationUnit)
			fParser.setSource((ICompilationUnit) unit);
		if (unit instanceof IClassFile)
			fParser.setSource((IClassFile) unit);
		fParser.setCompilerOptions(getCompilerOptions(unit));
		CompilationUnit result = (CompilationUnit) fParser.createAST(null);
		return result;
	}

	private Map getCompilerOptions(IJavaElement element) {
		IJavaProject project = element.getJavaProject();
		Map options = project.getOptions(true);
		for (Iterator iter = options.keySet().iterator(); iter.hasNext();) {
			String key = (String) iter.next();
			String value = (String) options.get(key);
			if ("error".equals(value) || "warning".equals(value)) { //$NON-NLS-1$//$NON-NLS-2$
				options.put(key, "ignore"); //$NON-NLS-1$
			}
		}
		options.put(JavaCore.COMPILER_TASK_TAGS, ""); //$NON-NLS-1$		
		return options;
	}
}