/*******************************************************************************
 * Copyright (c) 2008 Mateusz Matela and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     Mateusz Matela - initial API and implementation
 *******************************************************************************/
package org.gsoc.eclipse.tostringgenerator.internal.dialogs;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.internal.corext.codemanipulation.CodeGenerationSettings;
import org.eclipse.jdt.internal.corext.util.Messages;
import org.eclipse.jdt.internal.ui.JavaPlugin;
import org.eclipse.jdt.internal.ui.JavaPluginImages;
import org.eclipse.jdt.internal.ui.dialogs.SourceActionDialog;
import org.eclipse.jdt.internal.ui.dialogs.StatusInfo;
import org.eclipse.jdt.internal.ui.javaeditor.CompilationUnitEditor;
import org.eclipse.jdt.internal.ui.viewsupport.BindingLabelProvider;
import org.eclipse.jdt.internal.ui.viewsupport.JavaElementImageProvider;
import org.eclipse.jdt.ui.JavaElementImageDescriptor;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.dialogs.IDialogSettings;
import org.eclipse.jface.fieldassist.IContentProposal;
import org.eclipse.jface.fieldassist.IContentProposalProvider;
import org.eclipse.jface.fieldassist.TextContentAdapter;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.resource.StringConverter;
import org.eclipse.jface.viewers.CheckStateChangedEvent;
import org.eclipse.jface.viewers.CheckboxTreeViewer;
import org.eclipse.jface.viewers.ICheckStateListener;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.layout.FormAttachment;
import org.eclipse.swt.layout.FormData;
import org.eclipse.swt.layout.FormLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Spinner;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.dialogs.ISelectionStatusValidator;
import org.eclipse.ui.fieldassist.ContentAssistCommandAdapter;
import org.gsoc.eclipse.tostringgenerator.IToStringTemplateParser;
import org.gsoc.eclipse.tostringgenerator.internal.GenerateToStringOperation;

/**
 * Dialog for the generate toString() action.
 */
public class GenerateToStringDialog extends SourceActionDialog {

	private static class GenerateToStringContentProvider implements
			ITreeContentProvider {

		private static final Object[] EMPTY = new Object[0];

		private Object[] fFields;
		private Object[] fMethods;
		private Object[] fInheritedFields;
		private Object[] fInheritedMethods;
		private Object[] mainNodes;
		private static final String fieldsNode = UIMessages.GenerateToStringDialog_nodeFields;
		private static final String methodsNode = UIMessages.GenerateToStringDialog_nodeMethods;
		private static final String inheritedFieldsNode = UIMessages.GenerateToStringDialog_nodeInheritedFields;
		private static final String inheritedMethodsNode = UIMessages.GenerateToStringDialog_nodeInheritedMethods;

		public GenerateToStringContentProvider(IVariableBinding[] fields,
				IVariableBinding[] inheritedFields, IMethodBinding[] methods,
				IMethodBinding[] inheritedMethods) {
			ArrayList nodes = new ArrayList();
			fFields = (Object[]) fields.clone();
			if (fFields.length > 0)
				nodes.add(fieldsNode);
			fInheritedFields = (Object[]) inheritedFields.clone();
			if (fInheritedFields.length > 0)
				nodes.add(inheritedFieldsNode);
			fMethods = (Object[]) methods.clone();
			if (fMethods.length > 0)
				nodes.add(methodsNode);
			fInheritedMethods = (Object[]) inheritedMethods.clone();
			if (fInheritedMethods.length > 0)
				nodes.add(inheritedMethodsNode);
			mainNodes = nodes.toArray();
		}

		private int getElementPosition(Object element, Object[] array) {
			for (int i = 0; i < array.length; i++) {
				if (array[i].equals(element)) {
					return i;
				}
			}
			return -1;
		}

		private Object[] getContainingArray(Object element) {
			if (element instanceof String)
				return mainNodes;
			if (element instanceof IVariableBinding) {
				if (getElementPosition(element, fFields) >= 0)
					return fFields;
				if (getElementPosition(element, fInheritedFields) >= 0)
					return fInheritedFields;
			}
			if (element instanceof IMethodBinding) {
				if (getElementPosition(element, fMethods) >= 0)
					return fMethods;
				if (getElementPosition(element, fInheritedMethods) >= 0)
					return fInheritedMethods;
			}
			return EMPTY;
		}

		public boolean canMoveDown(Object element) {
			Object[] array = getContainingArray(element);
			int position = getElementPosition(element, array);
			return position != -1 && position != array.length - 1;
		}

		public boolean canMoveUp(Object element) {
			return getElementPosition(element, getContainingArray(element)) > 0;
		}

		public void down(Object element, CheckboxTreeViewer tree) {
			move(element, tree, 1);
		}

		public void up(Object element, CheckboxTreeViewer tree) {
			move(element, tree, -1);
		}

		private void move(Object element, CheckboxTreeViewer tree, int direction) {
			Object[] array = getContainingArray(element);
			int position = getElementPosition(element, array);
			Object temp = array[position];
			array[position] = array[position + direction];
			array[position + direction] = temp;
			tree.setSelection(new StructuredSelection(element));
			tree.refresh();
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see
		 * org.eclipse.jface.viewers.ITreeContentProvider#getChildren(java.lang
		 * .Object)
		 */
		public Object[] getChildren(Object parentElement) {
			if (parentElement == fieldsNode)
				return fFields;
			if (parentElement == inheritedFieldsNode)
				return fInheritedFields;
			if (parentElement == methodsNode)
				return fMethods;
			if (parentElement == inheritedMethodsNode)
				return fInheritedMethods;
			return EMPTY;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see
		 * org.eclipse.jface.viewers.ITreeContentProvider#getParent(java.lang
		 * .Object)
		 */
		public Object getParent(Object element) {
			Object[] array = getContainingArray(element);
			if (array == fFields)
				return fieldsNode;
			if (array == fInheritedFields)
				return inheritedFieldsNode;
			if (array == fMethods)
				return methodsNode;
			if (array == fInheritedMethods)
				return inheritedMethodsNode;
			return null;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see
		 * org.eclipse.jface.viewers.ITreeContentProvider#hasChildren(java.lang
		 * .Object)
		 */
		public boolean hasChildren(Object element) {
			if (element instanceof String)
				return true;
			return false;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see
		 * org.eclipse.jface.viewers.IStructuredContentProvider#getElements(
		 * java.lang.Object)
		 */
		public Object[] getElements(Object inputElement) {
			return mainNodes;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.eclipse.jface.viewers.IContentProvider#dispose()
		 */
		public void dispose() {
			// do nothing
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see
		 * org.eclipse.jface.viewers.IContentProvider#inputChanged(org.eclipse
		 * .jface.viewers.Viewer, java.lang.Object, java.lang.Object)
		 */
		public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {
			// do nothing
		}
	}

	private static class GenerateToStringLabelProvider extends
			BindingLabelProvider {
		public Image getImage(Object element) {
			ImageDescriptor descriptor = null;
			if (element == GenerateToStringContentProvider.fieldsNode
					|| element == GenerateToStringContentProvider.inheritedFieldsNode)
				descriptor = JavaPluginImages.DESC_FIELD_PUBLIC;
			if (element == GenerateToStringContentProvider.methodsNode
					|| element == GenerateToStringContentProvider.inheritedMethodsNode)
				descriptor = JavaPluginImages.DESC_MISC_PUBLIC;
			if (descriptor != null) {
				descriptor = new JavaElementImageDescriptor(descriptor, 0,
						JavaElementImageProvider.BIG_SIZE);
				return JavaPlugin.getImageDescriptorRegistry().get(descriptor);
			}
			return super.getImage(element);
		}
	}

	private static class GenerateToStringValidator implements
			ISelectionStatusValidator {

		private int fNumFields;
		private int fNumMethods;

		public GenerateToStringValidator(int fields, int methods) {
			fNumFields = fields;
			fNumMethods = methods;
		}

		public IStatus validate(Object[] selection) {
			int countFields = 0, countMethods = 0;
			for (int index = 0; index < selection.length; index++) {
				if (selection[index] instanceof IVariableBinding)
					countFields++;
				else if (selection[index] instanceof IMethodBinding)
					countMethods++;
			}

			return new StatusInfo(IStatus.INFO, Messages.format(
					UIMessages.GenerateToStringDialog_fieldsAndMethodsSelected,
					new String[] { String.valueOf(countFields),
							String.valueOf(fNumFields),
							String.valueOf(countMethods),
							String.valueOf(fNumMethods) }));
		}
	}

	private class ToStringTemplatesDialog extends Dialog {

		private class ToStringTemplateProposalProvider implements
				IContentProposalProvider {
			private class Proposal implements IContentProposal {
				final private String proposal;

				public Proposal(String proposal) {
					this.proposal = proposal;
				}

				public String getContent() {
					for (int i = 1; i < proposal.length(); i++) {
						if (proposal.substring(0, i).equals(
								latestContents.substring(latestPosition - i,
										latestPosition))) {
							return proposal.substring(i);
						}
					}
					return proposal;
				}

				public int getCursorPosition() {
					return proposal.length();
				}

				public String getDescription() {
					return (String) parser.getTokenDescriptions().get(proposal);
				}

				public String getLabel() {
					return proposal;
				}
			}

			private IContentProposal[] proposals;
			private String latestContents;
			private int latestPosition;

			public IContentProposal[] getProposals(String contents, int position) {
				if (proposals == null) {
					List proposalsList = new ArrayList();
					String[] tokens = parser.getTokens();
					for (int i = 0; i < tokens.length; i++) {
						proposalsList.add(new Proposal(tokens[i]));
					}
					proposals = (IContentProposal[]) proposalsList
							.toArray(new IContentProposal[0]);
				}
				this.latestContents = contents;
				this.latestPosition = position;
				return proposals;
			}

		}

		private final int ADD_BUTTON = IDialogConstants.CLIENT_ID + 1;

		private final int REMOVE_BUTTON = IDialogConstants.CLIENT_ID + 2;

		private final int APPLY_BUTTON = IDialogConstants.CLIENT_ID + 3;

		private Text templateTextControl;

		private Combo templateNameControl;

		private IToStringTemplateParser parser;

		private List templateNames;
		private List templates;
		private int selectedTemplate;

		private boolean controlsRefreshing = false;

		protected ToStringTemplatesDialog(Shell parentShell,
				IToStringTemplateParser parser) {
			super(parentShell);
			this.parser = parser;
		}

		protected Control createDialogArea(Composite parent) {
			getShell().setText(UIMessages.GenerateToStringDialog_manageTemplates);

			Composite composite = (Composite) super.createDialogArea(parent);

			Composite dialogSubArea = new Composite(composite, SWT.NONE);
			GridLayout gl = new GridLayout(3, false);
			dialogSubArea.setLayout(gl);
			GridData gd = new GridData(SWT.FILL, SWT.FILL, true, true);
			dialogSubArea.setLayoutData(gd);

			templateNameControl = new Combo(dialogSubArea, SWT.NONE);
			templateNameControl.setLayoutData(new GridData(SWT.FILL, SWT.FILL,
					true, true, 3, 1));
			templateNameControl.addSelectionListener(new SelectionListener() {

				public void widgetDefaultSelected(SelectionEvent e) {
					// do nothing
				}

				public void widgetSelected(SelectionEvent e) {
					if (templateNameControl.getSelectionIndex() >= 0) {
						selectedTemplate = templateNameControl
								.getSelectionIndex();
						templateTextControl.setText((String) templates
								.get(selectedTemplate));
					}
				}

			});

			templateTextControl = new Text(dialogSubArea, SWT.BORDER
					| SWT.MULTI | SWT.H_SCROLL);
			new ContentAssistCommandAdapter(templateTextControl,
					new TextContentAdapter(),
					new ToStringTemplateProposalProvider(), null,
					new char[] { '$' });

			gd = new GridData(SWT.FILL, SWT.FILL, true, true, 2, 3);
			gd.heightHint = 60;
			gd.widthHint = 300;
			templateTextControl.setLayoutData(gd);

			createButton(dialogSubArea, ADD_BUTTON, UIMessages.GenerateToStringDialog_new, false);
			createButton(dialogSubArea, REMOVE_BUTTON, UIMessages.GenerateToStringDialog_remove, false);

			gl.numColumns = 3;

			templateNames = new ArrayList(Arrays.asList(getTemplateNames()));
			selectedTemplate = getSelectedTemplate();
			templates = new ArrayList(Arrays
					.asList(getTemplates(getDialogSettings())));
			refreshControls();

			templateNameControl.addModifyListener(new ModifyListener() {
				private boolean selfRefreshing = false;

				public void modifyText(ModifyEvent e) {
					if (selectedTemplate >= 0
							&& !controlsRefreshing
							&& !selfRefreshing
							&& (templateNameControl.getSelectionIndex() == -1 || !templateNameControl
									.getText()
									.equals(
											templateNames
													.get(templateNameControl
															.getSelectionIndex())))) {
						templateNames.set(selectedTemplate, templateNameControl
								.getText());
						Point textSelection = templateNameControl
								.getSelection();
						selfRefreshing = true;
						templateNameControl.setItem(selectedTemplate,
								templateNameControl.getText());
						templateNameControl.select(selectedTemplate);
						templateNameControl.setSelection(textSelection);
						selfRefreshing = false;
						getButton(APPLY_BUTTON).setEnabled(true);
					}
				}
			});
			templateTextControl.addModifyListener(new ModifyListener() {
				public void modifyText(ModifyEvent e) {
					if (selectedTemplate >= 0
							&& !templateTextControl.getText().equals(
									templates.get(selectedTemplate))) {
						templates.set(selectedTemplate, templateTextControl
								.getText());
						getButton(APPLY_BUTTON).setEnabled(true);
					}
				}
			});

			return composite;
		}

		protected void createButtonsForButtonBar(Composite parent) {
			super.createButtonsForButtonBar(parent);
			createButton(parent, APPLY_BUTTON, UIMessages.GenerateToStringDialog_apply, true)
					.setEnabled(false);
		}

		private void applyChanges() {
			getDialogSettings().put(
					ToStringGenerationSettings.SETTINGS_TEMPLATE_NAMES,
					(String[]) templateNames.toArray(new String[0]));
			getDialogSettings().put(
					ToStringGenerationSettings.SETTINGS_TEMPLATES,
					(String[]) templates.toArray(new String[0]));
			getButton(APPLY_BUTTON).setEnabled(false);
		}

		protected void buttonPressed(int buttonId) {
			switch (buttonId) {
			case APPLY_BUTTON:
				applyChanges();
				break;
			case IDialogConstants.OK_ID:
				applyChanges();
				close();
				break;
			case IDialogConstants.CANCEL_ID:
				close();
				break;
			case ADD_BUTTON:
				templateNames.add(UIMessages.GenerateToStringDialog_newTemplate);
				templates.add(IToStringTemplateParser.DEFAULT_TEMPLATE);
				selectedTemplate = templateNames.size() - 1;
				getButton(APPLY_BUTTON).setEnabled(true);
				refreshControls();
				break;
			case REMOVE_BUTTON:
				if (templateNames.size() > 0) {
					templateNames.remove(selectedTemplate);
					templates.remove(selectedTemplate);
				}
				if (selectedTemplate >= templateNames.size())
					selectedTemplate = templateNames.size() - 1;
				getButton(APPLY_BUTTON).setEnabled(true);
				refreshControls();
				break;
			}
		}

		public void refreshControls() {
			controlsRefreshing = true;
			templateNameControl.setItems((String[]) templateNames
					.toArray(new String[0]));
			if (templateNames.size() > 0) {
				templateNameControl.select(selectedTemplate);
				templateTextControl.setText((String) templates
						.get(selectedTemplate));
			} else {
				templateTextControl.setText(""); //$NON-NLS-1$
			}
			controlsRefreshing = false;
		}
	}

	public static class ToStringGenerationSettings extends
			CodeGenerationSettings {

		private static final String SETTINGS_SELECTED_TEMPLATE = "ToStringTemplateSelected"; //$NON-NLS-1$

		private static final String SETTINGS_STRINGSTYLE = "StringStyle"; //$NON-NLS-1$

		private static final String SETTINGS_SKIPNULLS = "SkipNull"; //$NON-NLS-1$

		private static final String SETTINGS_IGNOREDEFAULT = "IgnoreDefault"; //$NON-NLS-1$

		private static final String SETTINGS_LIMITELEMENTS = "LimitElements"; //$NON-NLS-1$

		private static final String SETTINGS_LIMITVALUE = "LimitValue"; //$NON-NLS-1$

		private static final String SETTINGS_TEMPLATE_NAMES = "ToStringTemplateNames"; //$NON-NLS-1$

		private static final String SETTINGS_TEMPLATES = "ToStringTemplates"; //$NON-NLS-1$	

		/**
		 * which template should be used to format the output of the toString()
		 * method?
		 */
		public int stringFormatTemplateNumber;

		/**
		 * what is the template (redundancy - this field can be determined
		 * basing on <code>GenerateToStringDialog.getTemplates()</code> and
		 * <code>stringFormatTemplateNumber</code>, but this way it's more
		 * convenient)
		 */
		public String stringFormatTemplate;

		/** what style of code should the toString() method have? */
		public int toStringStyle;

		/** should the toString() method skip null values? */
		public boolean skipNulls;

		/**
		 * should the toString() method use its own way to show elements of an
		 * array?
		 */
		public boolean customArrayToString;

		/**
		 * should the toString() limit maximum number of elements of
		 * arrays/Collections to be listed?
		 */
		public boolean limitElements;

		/** what is the maximum number of elements in array/Collection to show? */
		public int limitValue;

		/** should blocks be forced in if/for/while statements? */
		public boolean useBlocks;

		/** can generated code use jdk 1.5 API? **/
		public boolean is50orHigher;

		/** can generated code use jdk 1.6 API? **/
		public boolean is60orHigher;

		protected boolean asBoolean(String string, boolean defaultValue) {
			if (string != null) {
				return StringConverter.asBoolean(string, defaultValue);
			}
			return defaultValue;
		}

		protected int asInt(String string, int defaultValue) {
			if (string != null) {
				return StringConverter.asInt(string, defaultValue);
			}
			return defaultValue;
		}

		public ToStringGenerationSettings(IDialogSettings dialogSettings) {
			limitElements = asBoolean(dialogSettings
					.get(SETTINGS_LIMITELEMENTS), false);
			customArrayToString = asBoolean(dialogSettings
					.get(SETTINGS_IGNOREDEFAULT), true);
			toStringStyle = asInt(dialogSettings.get(SETTINGS_STRINGSTYLE), 0);
			limitValue = asInt(dialogSettings.get(SETTINGS_LIMITVALUE), 10);
			skipNulls = asBoolean(dialogSettings.get(SETTINGS_SKIPNULLS), false);
			stringFormatTemplateNumber = asInt(dialogSettings
					.get(SETTINGS_SELECTED_TEMPLATE), 0);
		}

		public ToStringGenerationSettings() {
			// no initialization
		}

		public void writeDialogSettings(IDialogSettings dialogSettings) {
			dialogSettings.put(SETTINGS_LIMITELEMENTS, limitElements);
			dialogSettings.put(SETTINGS_IGNOREDEFAULT, customArrayToString);
			dialogSettings.put(SETTINGS_STRINGSTYLE, toStringStyle);
			dialogSettings.put(SETTINGS_LIMITVALUE, limitValue);
			dialogSettings.put(SETTINGS_SKIPNULLS, skipNulls);
			dialogSettings.put(SETTINGS_SELECTED_TEMPLATE,
					stringFormatTemplateNumber);
		}
	}

	private ToStringGenerationSettings fGenerationSettings;

	private static final int DOWN_BUTTON = IDialogConstants.CLIENT_ID + 2;

	private static final int UP_BUTTON = IDialogConstants.CLIENT_ID + 1;

	protected Button[] fButtonControls;

	boolean[] fButtonsEnabled;

	private static final int DOWN_INDEX = 1;

	private static final int UP_INDEX = 0;

	private Combo formatCombo;

	private IDialogSettings fDialogSettings;

	public ToStringGenerationSettings getGenerationSettings() {
		return fGenerationSettings;
	}

	public static String[] getTemplates(IDialogSettings dialogSettings) {
		String[] result = dialogSettings
				.getArray(ToStringGenerationSettings.SETTINGS_TEMPLATES);
		if (result != null && result.length > 0)
			return result;
		return new String[] { IToStringTemplateParser.DEFAULT_TEMPLATE };
	}

	public String[] getTemplateNames() {
		String[] result = getDialogSettings().getArray(
				ToStringGenerationSettings.SETTINGS_TEMPLATE_NAMES);
		if (result != null && result.length > 0)
			return result;
		return new String[] { UIMessages.GenerateToStringDialog_defaultTemplateName };
	}

	public int getSelectedTemplate() {
		try {
			return getDialogSettings().getInt(
					ToStringGenerationSettings.SETTINGS_SELECTED_TEMPLATE);
		} catch (NumberFormatException e) {
			return 0;
		}
	}

	public void setSelectedTemplate(int templateNumber) {
		getDialogSettings().put(
				ToStringGenerationSettings.SETTINGS_SELECTED_TEMPLATE,
				templateNumber);
	}

	public GenerateToStringDialog(Shell shell, CompilationUnitEditor editor,
			IType type, IVariableBinding[] fields,
			IVariableBinding[] inheritedFields,
			IVariableBinding[] selectedFields, IMethodBinding[] methods,
			IMethodBinding[] inheritededMethods) throws JavaModelException {
		super(shell, new BindingLabelProvider(),
				new GenerateToStringContentProvider(fields, inheritedFields,
						methods, inheritededMethods), editor, type, false);
		setEmptyListMessage(UIMessages.GenerateToStringDialog_noFieldsNorMethods);

		List selected = new ArrayList(Arrays.asList(selectedFields));
		if (selectedFields.length == fields.length && selectedFields.length > 0)
			selected.add(getContentProvider().getParent(selectedFields[0]));
		setInitialElementSelections(selected);

		setTitle(UIMessages.GenerateToStringDialog_dialogTitle);
		setMessage(UIMessages.GenerateToStringDialog_selectFieldsAndMethods);
		setValidator(new GenerateToStringValidator(fields.length
				+ inheritedFields.length, methods.length
				+ inheritededMethods.length));
		setSize(60, 18);
		setInput(new Object());

		IDialogSettings dialogSettings = JavaPlugin.getDefault()
				.getDialogSettings();
		String sectionId = "SourceActionDialog.methods"; //$NON-NLS-1$
		fDialogSettings = dialogSettings.getSection(sectionId);
		if (fDialogSettings == null) {
			fDialogSettings = dialogSettings.addNewSection(sectionId);
		}
		fGenerationSettings = new ToStringGenerationSettings(
				getDialogSettings());
	}

	public IDialogSettings getDialogSettings() {
		return fDialogSettings;
	}

	/**
	 * {@inheritDoc}
	 */
	public boolean close() {
		boolean result = super.close();

		fGenerationSettings.writeDialogSettings(getDialogSettings());

		fGenerationSettings.stringFormatTemplate = getTemplates(getDialogSettings())[fGenerationSettings.stringFormatTemplateNumber];

		fGenerationSettings.createComments = getGenerateComment();

		return result;
	}

	public Object[] getResult() {
		Object[] oldResult = super.getResult();
		List newResult = new ArrayList();
		for (int i = 0; i < oldResult.length; i++) {
			if (!(oldResult[i] instanceof String)) {
				newResult.add(oldResult[i]);
			}
		}
		return newResult.toArray();
	}

	protected CheckboxTreeViewer createTreeViewer(Composite parent) {
		CheckboxTreeViewer treeViewer = super.createTreeViewer(parent);

		treeViewer.setLabelProvider(new GenerateToStringLabelProvider());

		treeViewer.expandToLevel(GenerateToStringContentProvider.fieldsNode, 1);

		treeViewer.addSelectionChangedListener(new ISelectionChangedListener() {
			public void selectionChanged(SelectionChangedEvent event) {
				IStructuredSelection selection = (IStructuredSelection) getTreeViewer()
						.getSelection();

				Object selected = selection.size() > 0 ? selection.toList()
						.get(0) : null;
				GenerateToStringContentProvider cp = (GenerateToStringContentProvider) getContentProvider();

				fButtonControls[UP_INDEX].setEnabled(cp.canMoveUp(selected));
				fButtonControls[DOWN_INDEX]
						.setEnabled(cp.canMoveDown(selected));
			}

		});
		treeViewer.addCheckStateListener(new ICheckStateListener() {
			public void checkStateChanged(CheckStateChangedEvent event) {
				getTreeViewer().setSubtreeChecked(event.getElement(),
						event.getChecked());
				getTreeViewer().setGrayed(event.getElement(), false);
				Object parentElement = ((ITreeContentProvider) (getTreeViewer()
						.getContentProvider())).getParent(event.getElement());
				if (parentElement != null) {
					Object[] siblings = ((ITreeContentProvider) (getTreeViewer()
							.getContentProvider())).getChildren(parentElement);
					int count = 0;
					for (int i = 0; i < siblings.length; i++) {
						if (getTreeViewer().getChecked(siblings[i]))
							count++;
					}
					if (count == 0)
						getTreeViewer().setGrayChecked(parentElement, false);
					else if (count == siblings.length) {
						getTreeViewer().setChecked(parentElement, true);
						getTreeViewer().setGrayed(parentElement, false);
					} else
						getTreeViewer().setGrayChecked(parentElement, true);
				}
				updateOKStatus();
			}

		});
		return treeViewer;
	}

	protected Composite createSelectionButtons(Composite composite) {
		Composite buttonComposite = super.createSelectionButtons(composite);

		GridLayout layout = new GridLayout();
		buttonComposite.setLayout(layout);

		createUpDownButtons(buttonComposite);

		layout.marginHeight = 0;
		layout.marginWidth = 0;
		layout.numColumns = 1;

		return buttonComposite;
	}

	protected void buttonPressed(int buttonId) {
		super.buttonPressed(buttonId);
		switch (buttonId) {
		case UP_BUTTON: {
			GenerateToStringContentProvider contentProvider = (GenerateToStringContentProvider) getTreeViewer()
					.getContentProvider();
			List selection = ((IStructuredSelection) getTreeViewer()
					.getSelection()).toList();
			if (selection.size() > 0)
				contentProvider.up(selection.get(0), getTreeViewer());
			updateOKStatus();
			break;
		}
		case DOWN_BUTTON: {
			GenerateToStringContentProvider contentProvider = (GenerateToStringContentProvider) getTreeViewer()
					.getContentProvider();
			List selection = ((IStructuredSelection) getTreeViewer()
					.getSelection()).toList();
			if (selection.size() > 0)
				contentProvider.down(selection.get(0), getTreeViewer());
			updateOKStatus();
			break;
		}
		}
	}

	protected void createUpDownButtons(Composite buttonComposite) {
		int numButtons = 2; // up, down
		fButtonControls = new Button[numButtons];
		fButtonsEnabled = new boolean[numButtons];
		fButtonControls[UP_INDEX] = createButton(buttonComposite, UP_BUTTON,
				UIMessages.GenerateToStringDialog_up, false);
		fButtonControls[DOWN_INDEX] = createButton(buttonComposite,
				DOWN_BUTTON, UIMessages.GenerateToStringDialog_down, false);
		boolean defaultState = false;
		fButtonControls[UP_INDEX].setEnabled(defaultState);
		fButtonControls[DOWN_INDEX].setEnabled(defaultState);
		fButtonsEnabled[UP_INDEX] = defaultState;
		fButtonsEnabled[DOWN_INDEX] = defaultState;
	}

	protected Composite createCommentSelection(Composite parentComposite) {
		Composite composite = super.createCommentSelection(parentComposite);

		Composite composite2 = new Composite(parentComposite, SWT.NONE);
		GridLayout layout = new GridLayout(3, false);
		layout.marginWidth = 0;
		composite2.setLayout(layout);
		composite2.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));

		final Label formatLabel = new Label(composite2, SWT.RIGHT);
		formatLabel.setText(UIMessages.GenerateToStringDialog_stringFormat);
		GridData gridData = new GridData(SWT.FILL, SWT.CENTER, false, false);
		formatLabel.setLayoutData(gridData);

		formatCombo = new Combo(composite2, SWT.READ_ONLY);
		formatCombo.setItems(getTemplateNames());
		formatCombo.select(fGenerationSettings.stringFormatTemplateNumber);
		formatCombo
				.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
		formatCombo.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				fGenerationSettings.stringFormatTemplateNumber = ((Combo) e.widget)
						.getSelectionIndex();
			}
		});

		final Button styleButton = new Button(composite2, SWT.NONE);
		styleButton.setText(UIMessages.GenerateToStringDialog_openTemplateManger);
		styleButton.setLayoutData(new GridData(SWT.RIGHT, SWT.FILL, false,
				false));
		styleButton.addSelectionListener(new SelectionListener() {
			public void widgetSelected(SelectionEvent e) {
				ToStringTemplatesDialog dialog = new ToStringTemplatesDialog(
						getShell(),
						GenerateToStringOperation
								.createTemplateParser(fGenerationSettings.toStringStyle));
				dialog.open();
				formatCombo.setItems(getTemplateNames());
				fGenerationSettings.stringFormatTemplateNumber = Math.min(
						fGenerationSettings.stringFormatTemplateNumber,
						formatCombo.getItemCount() - 1);
				formatCombo
						.select(fGenerationSettings.stringFormatTemplateNumber);
			}

			public void widgetDefaultSelected(SelectionEvent e) {
				// do nothing
			}
		});

		final Label styleLabel = new Label(composite2, SWT.RIGHT);
		styleLabel.setText(UIMessages.GenerateToStringDialog_codeStyle);
		gridData = new GridData(SWT.FILL, SWT.CENTER, false, false);
		styleLabel.setLayoutData(gridData);

		final Combo styleCombo = new Combo(composite2, SWT.READ_ONLY);
		styleCombo.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false,
				2, 1));
		styleCombo.setItems(GenerateToStringOperation.refreshStyles());
		styleCombo.select(Math.min(fGenerationSettings.toStringStyle,
				styleCombo.getItemCount() - 1));
		styleCombo.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				fGenerationSettings.toStringStyle = ((Combo) e.widget)
						.getSelectionIndex();
			}
		});

		final Button skipNullsButton = new Button(composite2, SWT.CHECK);
		skipNullsButton.setText(UIMessages.GenerateToStringDialog_skipNulls);
		skipNullsButton.setSelection(fGenerationSettings.skipNulls);
		gridData = new GridData(SWT.FILL, SWT.FILL, true, true, 3, 1);
		skipNullsButton.setLayoutData(gridData);
		skipNullsButton.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent event) {
				fGenerationSettings.skipNulls = ((Button) event.widget)
						.getSelection();
			}
		});

		final Button arrayButton = new Button(composite2, SWT.CHECK);
		arrayButton.setText(UIMessages.GenerateToStringDialog_ignoreArraysToString);
		arrayButton.setSelection(fGenerationSettings.customArrayToString);
		gridData = new GridData(SWT.FILL, SWT.FILL, true, true, 3, 1);
		arrayButton.setLayoutData(gridData);
		arrayButton.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				fGenerationSettings.customArrayToString = ((Button) e.widget)
						.getSelection();
			}
		});

		Composite limitComposite = new Composite(composite2, SWT.NONE);
		limitComposite.setLayout(new FormLayout());
		gridData = new GridData(SWT.FILL, SWT.FILL, true, true, 3, 1);
		limitComposite.setLayoutData(gridData);

		final Button limitButton = new Button(limitComposite, SWT.CHECK);
		limitButton
				.setText(UIMessages.GenerateToStringDialog_limitItems);
		limitButton.setSelection(fGenerationSettings.limitElements);
		FormData formData = new FormData();
		limitButton.setLayoutData(formData);
		limitButton.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				fGenerationSettings.limitElements = ((Button) e.widget)
						.getSelection();
			}
		});

		final Spinner limitSpinner = new Spinner(limitComposite, SWT.NONE);
		limitSpinner.setMinimum(0);
		limitSpinner.setSelection(fGenerationSettings.limitValue);
		formData = new FormData();
		formData.left = new FormAttachment(limitButton, 2);
		limitSpinner.setLayoutData(formData);
		limitSpinner.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				fGenerationSettings.limitValue = ((Spinner) e.widget)
						.getSelection();
			}
		});
		return composite;
	}

	protected void configureShell(Shell shell) {
		super.configureShell(shell);
		PlatformUI
				.getWorkbench()
				.getHelpSystem()
				.setHelp(shell,
						"org.gsoc.eclipse.tostringgenerator.to_string_tree_selection_dialog_context"); //$NON-NLS-1$
	}

}